@@ -32,12 +32,16 @@ __all__ = [
     "fastjet_clusters",
     "leaf_masks",
     "centroid_prune",
+    "color_singlets",
+    "clusters",
 ]
 
 
 DistFunc = ty.Callable[
     [gcl.MomentumArray, gcl.MomentumArray], base.DoubleVector
 ]
+MaskType = ty.Union[gcl.MaskArray, gcl.MaskGroup]
+MaskGeneric = ty.TypeVar("MaskGeneric", gcl.MaskGroup, gcl.MaskArray, MaskType)
 
 
 def fastjet_clusters(
@@ -82,7 +86,7 @@ def fastjet_clusters(
 
     Returns
     -------
-    clusters : list[MaskArray]
+    list[MaskArray]
         List containing masks over the input data for each jet
         clustering, in order of descending :math:`p_T`.
 
@@ -140,7 +144,7 @@ def find_vertex(
 
     Returns
     -------
-    vertices : ndarray[int32]
+    ndarray[int32]
         List the vertex ids which match the passed incoming and outgoing
         pdg codes.
 
@@ -203,16 +207,17 @@ def find_vertex(
 
 
 def vertex_descendants(adj: gcl.AdjacencyList, vertex: int) -> gcl.MaskArray:
-    """Returns a ``MaskArray`` to select particles which descend from a
-    given interaction vertex. This mask includes the parent vertex.
+    """Returns a ``MaskArray`` to select edges which descend from a
+    given interaction vertex.
 
     :group: select
 
     .. versionadded:: 0.1.0
 
     .. versionchanged:: 0.2.7
-       In the edge case of no descendants, now returns a mask with just
-       the parent, rather than raising an unhandled ``IndexError``.
+       In the edge case of no descendants, now returns a mask with
+       identifying edges whose source is the vertex, rather than raising
+       an unhandled ``IndexError``.
 
     Parameters
     ----------
@@ -223,7 +228,7 @@ def vertex_descendants(adj: gcl.AdjacencyList, vertex: int) -> gcl.MaskArray:
 
     Returns
     -------
-    mask : MaskArray
+    MaskArray
         Boolean mask over the graphicle objects associated with the
         passed AdjacencyList.
     """
@@ -232,12 +237,40 @@ def vertex_descendants(adj: gcl.AdjacencyList, vertex: int) -> gcl.MaskArray:
         return gcl.MaskArray(adj.edges["out"] == vertex)
     sparse = adj._sparse_signed
     vertex = sparse.row[vertex_mask][0]
-    bft = breadth_first_tree(sparse, vertex)
+    bft = breadth_first_tree(adj._sparse_csr, vertex)
     mask = np.isin(sparse.row, bft.indices)
-    mask[sparse.row == vertex] = True  # include parent vertex
+    mask[sparse.row == vertex] = True  # include edges directly from parent
     return gcl.MaskArray(mask)
 
 
+def edge_descendants(
+    adj: gcl.AdjacencyList, edge: gcl.VertexPair
+) -> gcl.MaskArray:
+    """Returns a ``MaskArray`` to select particles which descend from a
+    given edge in the DAG. This mask includes the parent edge.
+
+    :group: select
+
+    .. versionadded:: 0.2.8
+
+    Parameters
+    ----------
+    adj : AdjacencyList
+        Topological structure of the graph.
+    edge : VertexPair or tuple[int, int]
+        The vertex id from which the descending edges are identified.
+
+    Returns
+    -------
+    MaskArray
+        Boolean mask over the graphicle objects associated with the
+        passed AdjacencyList.
+    """
+    desc = vertex_descendants(adj, edge[1])
+    desc[adj == edge] = True
+    return desc
+
+
 def hadron_vertices(
     adj: gcl.AdjacencyList,
     status: gcl.StatusArray,
@@ -257,7 +290,7 @@ def hadron_vertices(
 
     Returns
     -------
-    vertex_ids : tuple[int]
+    tuple[int, ...]
         Indices of the hadronisation vertices in the generation DAG,
         returned in no particular order.
     """
@@ -328,7 +361,7 @@ def _partition_vertex(
 
     Returns
     -------
-    filtered_mask : MaskArray
+    MaskArray
         Input ``MaskArray``, filtered to remove background incident on
         the same hadronisation vertex.
     """
@@ -345,9 +378,9 @@ def _partition_vertex(
 
 def partition_descendants(
     graph: gcl.Graphicle,
-    hier: gcl.MaskGroup,
+    hier: gcl.MaskGroup[MaskGeneric],
     pt_exp: float = -0.1,
-) -> gcl.MaskGroup:
+) -> gcl.MaskGroup[MaskGeneric]:
     """Partitions the final state descendants with mixed hard partonic
     heritage, by aligning them with their nearest ancestor.
 
@@ -372,7 +405,7 @@ def partition_descendants(
 
     Returns
     -------
-    hier_parted : MaskGroup
+    MaskGroup
         Same nested tree structure as input, but with the final
         state hadrons partitioned to their nearest hard parton ancestor.
     """
@@ -398,7 +431,9 @@ def partition_descendants(
     return hier
 
 
-def _hard_ancestor_matrix(hard_desc: gcl.MaskGroup) -> base.BoolVector:
+def _hard_ancestor_matrix(
+    hard_desc: gcl.MaskGroup[MaskType],
+) -> base.BoolVector:
     """Returns a directed adjacency matrix for the hard process.
 
     Parameters
@@ -411,7 +446,7 @@ def _hard_ancestor_matrix(hard_desc: gcl.MaskGroup) -> base.BoolVector:
 
     Returns
     -------
-    hard_matrix : ndarray[bool_]
+    ndarray[bool_]
         Square directed adjacency matrix, linking partons in the hard
         process to their descendants. Rows represent each hard parton,
         and columns indicate which hard partons descend from it
@@ -438,7 +473,7 @@ def _hard_parent_matrix(hard_matrix: base.BoolVector) -> base.BoolVector:
 
     Returns
     -------
-    hard_parent_matrix : ndarray[bool_]
+    ndarray[bool_]
         Directed adjacency matrix representing links from parents to
         direct (first generation) children. Same as ``hard_matrix``, but
         with grandchildren *etc.* removed.
@@ -479,7 +514,7 @@ def _hard_parent_dict(
 
     Returns
     -------
-    parent_dict : dict[str, tuple[str, ...]]
+    dict[str, tuple[str, ...]]
         Mapping from parents to children in the hard process. Keys are
         parent names, and values are tuples of children names.
     """
@@ -493,7 +528,7 @@ def _hard_parent_dict(
 
 
 def _flat_hierarchy(
-    hard_desc: gcl.MaskGroup,
+    hard_desc: gcl.MaskGroup[gcl.MaskArray],
 ) -> ty.Dict[str, ty.Tuple[str, ...]]:
     """Produces a mapping between partons to their direct children in
     the hard process. This mapping is flat, *ie.* there is no nesting
@@ -512,7 +547,7 @@ def _flat_hierarchy(
 
     Returns
     -------
-    parent_dict : dict[str, tuple[str, ...]]
+    dict[str, tuple[str, ...]]
         Mapping from parents to children in the hard process. Keys are
         parent names, and values are tuples of children names.
 
@@ -557,7 +592,7 @@ def hard_descendants(
     target: ty.Optional[ty.Iterable[int]] = None,
     sign_sensitive: bool = False,
     strict: bool = True,
-) -> gcl.MaskGroup:
+) -> gcl.MaskGroup[gcl.MaskArray]:
     """Returns a ``MaskGroup`` instance to select particle descendants
     of ``target`` hard partons (by PDG code).
 
@@ -593,7 +628,7 @@ def hard_descendants(
 
     Returns
     -------
-    descendant_masks : MaskGroup
+    MaskGroup[MaskArray]
         Collection of masks over the event indicating the descendants
         of partons from within the hard process.
 
@@ -623,14 +658,17 @@ def hard_descendants(
         hard_graph = hard_graph[target_mask]
     pdg_keys = _pdgs_to_keys(hard_graph.pdg)
     pcl_out_vtxs = map(int, hard_graph.edges["out"])
-    vtx_descs = map(vertex_descendants, it.repeat(graph.adj), pcl_out_vtxs)
-    return gcl.MaskGroup(cl.OrderedDict(zip(pdg_keys, vtx_descs)), agg_op="or")
+    descs = map(vertex_descendants, it.repeat(graph.adj), pcl_out_vtxs)
+    group = gcl.MaskGroup(cl.OrderedDict(zip(pdg_keys, descs)), agg_op="or")
+    for key, idx in zip(pdg_keys, np.flatnonzero(hard_mask), strict=True):
+        group[key][idx] = True
+    return group
 
 
 def hierarchy(
     graph: gcl.Graphicle,
-    desc: ty.Optional[gcl.MaskGroup] = None,
-) -> gcl.MaskGroup:
+    desc: ty.Optional[gcl.MaskGroup[gcl.MaskArray]] = None,
+) -> gcl.MaskGroup[MaskType]:
     """Composite ``MaskGroup`` of ``MaskGroup`` instances, representing
     the partons of the hard process and their descendants. Uses a tree
     structure, such that partons which are descendants of other hard
@@ -645,14 +683,14 @@ def hierarchy(
     graph : Graphicle
         The event to be parsed. Must include ``status``, ``edges``,
         ``pmu``, and ``pdg`` data.
-    desc : MaskGroup, optional
+    desc : MaskGroup[MaskArray], optional
         If the masks for all partons in the hard process have already
         been computed, these may be passed here to save performing the
         computation again.
 
     Returns
     -------
-    hierarchy : MaskGroup
+    MaskGroup[MaskGroup | MaskArray]
         Nested composite of ``MaskGroup`` instances, representing the
         hierarchical structure of the hard process, and the descendants
         of the hard process partons throughout the shower. Nested
@@ -716,7 +754,7 @@ def hierarchy(
         desc = hard_descendants(graph)
     else:
         desc = desc.copy()
-    hard_desc: gcl.MaskGroup = desc[hard_mask][list(names)]  # type: ignore
+    hard_desc = desc[hard_mask][list(names)]
     hard = _flat_hierarchy(hard_desc)
     keys = set(hard.keys())
     vals = set(it.chain.from_iterable(hard.values()))
@@ -728,14 +766,6 @@ def hierarchy(
         for key, idx in zip(hard_final_keys, hard_final_idxs):
             desc[key].data[idx] = True
     masks = _make_tree(hard, roots, hard, desc, graph.adj)
-    leaf_iters = it.starmap(
-        fn.partial(_leaf_mask_iter, exclude_latent=False), masks.items()
-    )
-    leaf_iter = it.chain.from_iterable(leaf_iters)
-    leaf_iter = filter(lambda kv: kv[0] == "latent", leaf_iter)
-    leaf_masks = map(op.itemgetter(1), leaf_iter)
-    for leaf_mask in leaf_masks:
-        leaf_mask[leaf_mask & graph.hard_mask] = False
     return masks
 
 
@@ -743,9 +773,9 @@ def _make_tree(
     flat: ty.Union[ty.Dict[str, ty.Any], ty.Tuple[str, ...]],
     roots: ty.Set[str],
     hard: ty.Dict[str, ty.Tuple[str, ...]],
-    desc: gcl.MaskGroup,
+    desc: gcl.MaskGroup[gcl.MaskArray],
     adj: gcl.AdjacencyList,
-) -> gcl.MaskGroup:
+) -> gcl.MaskGroup[MaskType]:
     """Recursive function to convert a flat representation of the hard
     process tree into a nested ``MaskGroup`` of ``MaskGroup`` object.
 
@@ -792,7 +822,7 @@ def _make_tree(
 
 def _leaf_mask_iter(
     branch_name: str,
-    branch: ty.Union[gcl.MaskGroup, gcl.MaskArray],
+    branch: MaskType,
     exclude_latent: bool = True,
 ) -> ty.Generator[ty.Tuple[str, gcl.MaskArray], None, None]:
     """Recursive function, traversing a branch of the nested mask tree
@@ -808,7 +838,7 @@ def _leaf_mask_iter(
             yield from _leaf_mask_iter(name, mask, exclude_latent)  # type: ignore
 
 
-def leaf_masks(mask_tree: gcl.MaskGroup) -> gcl.MaskGroup:
+def leaf_masks(mask_tree: gcl.MaskGroup) -> gcl.MaskGroup[gcl.MaskArray]:
     """Find the leaves of the hard process, when organised into a
     hierarchical tree from ``hierarchy()``.
 
@@ -824,7 +854,7 @@ def leaf_masks(mask_tree: gcl.MaskGroup) -> gcl.MaskGroup:
 
     Returns
     -------
-    leaves : MaskGroup
+    MaskGroup
         Flat ``MaskGroup`` of only the leaves of ``mask_tree``.
     """
     mask_group = gcl.MaskGroup(agg_op="or")  # type: ignore
@@ -833,7 +863,7 @@ def leaf_masks(mask_tree: gcl.MaskGroup) -> gcl.MaskGroup:
     return mask_group
 
 
-def any_overlap(masks: gcl.MaskGroup) -> bool:
+def any_overlap(masks: gcl.MaskGroup[MaskType]) -> bool:
     """Given a ``MaskGroup`` object, checks if any of the masks overlap
     with each other.
 
@@ -848,7 +878,7 @@ def any_overlap(masks: gcl.MaskGroup) -> bool:
 
     Returns
     -------
-    any_overlap : bool
+    bool
         ``True`` if at least two ``MaskArrays`` in ``masks`` have at
         least one ``True`` element in the same location.
     """
@@ -891,7 +921,7 @@ def centroid_prune(
 
     Returns
     -------
-    prune_mask : MaskArray
+    MaskArray
         Mask which retains only the particles within ``radius`` of the
         centroid.
     """
@@ -911,3 +941,166 @@ def centroid_prune(
         return gcl.MaskArray(is_within)
     event_mask[mask] = is_within
     return gcl.MaskArray(event_mask)
+
+
+def color_singlets(
+    leaves: gcl.MaskGroup[gcl.MaskArray],
+    status: gcl.StatusArray,
+    color: gcl.ColorArray,
+    invert: bool = False,
+) -> ty.List[ty.Tuple[str, ...]]:
+    """Identifies groups of outgoing partons from the hard process which
+    form color singlets.
+
+    :group: select
+
+    .. versionadded:: 0.2.8
+
+    Parameters
+    ----------
+    leaves : MaskGroup[MaskArray]
+        Innermost nested ``MaskArray`` instances, obtained from calling
+        ``hierarchy()`` followed by ``leaf_masks()`` on a ``Graphicle``
+        object.
+    status : StatusArray
+        The status codes for the event.
+    color : ColorArray
+        The color / anticolor pair values for the event.
+    invert : bool
+        If ``True``, will invert the operation to find all colored
+        particles in the hard process which are not grouped with other
+        partons in the hard-process to form a color-singlet. Default is
+        ``False``.
+
+    Returns
+    -------
+    list[tuple[str, ...]]
+        Tuples of keys, grouping the color-connected partons in the hard
+        process.
+
+    Notes
+    -----
+    This function checks for color singlets formed from two color
+    triplets (quarks), or two color triplets plus a color octet (gluon).
+    It does not check for color singlets formed from three colour
+    triplets. This functionality *may* be added in future.
+    """
+    hard_mask = status.hard_mask["outgoing"]
+    masks = map(op.and_, it.repeat(hard_mask), leaves.values())
+    colors_ = map(it.compress, it.repeat(color), masks)
+    colors = tuple(it.chain.from_iterable(colors_))
+    colored_mask = tuple(map(op.ne, it.repeat((0, 0)), colors))
+    colored_only = it.compress(colors, colored_mask)
+    keys = tuple(it.compress(leaves.keys(), colored_mask))
+    named_colors = tuple(zip(keys, colored_only, strict=True))  # type: ignore
+    combos = it.chain(
+        it.combinations(named_colors, 2), it.combinations(named_colors, 3)
+    )
+    singlets = []
+    for names, color_elems in it.starmap(zip, combos):
+        color_set = set(map(op.attrgetter("color"), color_elems))
+        anticolor_set = set(map(op.attrgetter("anticolor"), color_elems))
+        if color_set == anticolor_set:
+            singlets.append(names)
+    if invert is True:
+        members = set(it.chain.from_iterable(singlets))
+        outcasts = set(keys).symmetric_difference(members)
+        return list(zip(outcasts))
+    return singlets
+
+
+def clusters(
+    graph: gcl.Graphicle, radius: float
+) -> gcl.MaskGroup[gcl.MaskArray]:
+    """Cluster and tag the final state particles in an event represented
+    by a ``Graphicle`` object. These clusters are formed by considering
+    the topology of the directed acyclic graph (DAG) generating the
+    event, tracking descendants of hard partons, and the momenta of the
+    hard partons compared against the final state particles.
+
+    :group: select
+
+    .. versionadded:: 0.2.8
+
+    The steps defining this algorithm are summarised:
+
+    * Find descendants of all hard partons within DAG
+    * Remove final state radiation from descendants
+    * Where descendants of multiple hard partons annihilate color with
+      each other, assign exclusive parentage of subsequent color neutral
+      particles to closest hard parton in the pseudorapidity-azimuth
+      (:math:`\\eta-\\phi`) plane
+    * Where background is used to annihilate color of hard parton
+      descendants, remove all final state particles beyond a distance
+      of ``radius`` from the position of the hard parton in the
+      :math:`\\eta-\\phi` plane
+
+    Parameters
+    ----------
+    graph : Graphicle
+        Full event record as a DAG.
+    radius : float
+        Radius in the :math:`\\eta-\\phi` plane defining the enclosed
+        clustering region around hard partons which are color-connected
+        to the underlying event. See notes for more information.
+
+    Returns
+    -------
+    MaskGroup[MaskArray]
+        Flat ``MaskGroup``, containing ``MaskArray`` instances which
+        reconstruct the hard partons from the final state particles.
+
+    Notes
+    -----
+    Hard partons may be color connected with each other, or with the
+    underlying event. *eg.* for a hierarchical clustering:
+
+    .. code-block:: text
+
+        MaskGroup(agg_op=OR)
+        ├── t
+        │   ├── b
+        │   └── W+
+        │       ├── c
+        │       └── s~
+        └── t~
+            ├── b~
+            └── W-
+                ├── s
+                └── c~
+
+
+    the quarks decaying from the W bosons are color-connected to each
+    other, as they form from a color-singlet. This means they will
+    almost certainly annihilate their color with each other during
+    hadronisation. However, the top quarks are color connected to the
+    underlying event, and therefore will almost certainly annihilate
+    their color with partons that do not descend from the hard process.
+    This is done via proxy of the bottom quark, which inherits its color
+    from the top. This results in background radiation in the
+    descendants tree of the bottom quark, spread over a wide region of
+    the :math:`\\eta-\\phi` plane. This is when the ``radius`` parameter
+    is applied, excluding all final state descendants whose distance
+    from the hard bottom quark exceeds the value passed, cleaning up the
+    signal.
+    """
+    hier_ = hierarchy(graph)
+    hier = partition_descendants(graph, hier_, -0.1)
+    leaves = leaf_masks(hier)
+    color_keys = color_singlets(leaves, graph.status, graph.color, True)
+    colored_leaves = map(
+        op.getitem, it.repeat(leaves), it.chain.from_iterable(color_keys)
+    )
+    colored_leaves, colored_leaves_ = it.tee(colored_leaves, 2)
+    hard_mask = graph.hard_mask["outgoing"]
+    parton_masks = map(op.and_, it.repeat(hard_mask), colored_leaves_)
+    parton_pmus = map(op.getitem, it.repeat(graph.pmu), parton_masks)
+    parton_centroids = map(op.attrgetter("eta", "phi"), parton_pmus)
+    for leaf, centroid in zip(colored_leaves, parton_centroids):
+        leaf[...] = centroid_prune(graph.pmu, radius, leaf, centroid)
+    hier.recursive_drop(inplace=True)
+    flat_hier = hier.flatten("rise")
+    flat_hier_final = map(op.itemgetter(graph.final), flat_hier.values())
+    return gcl.MaskGroup(
+        cl.OrderedDict(zip(flat_hier.keys(), flat_hier_final)), agg_op="or"
+    )
